﻿// SitemapWiz v 1.0 (c) 2022 by Enrico Savazzi. Do not remove or alter this copyright notice.

using Nancy.Json;
using System;
using System.Text;

namespace SitemapWriter1 
{
 
    internal class Program
    {
         static void Main(string[] args)
        {
            // read the app settings
            MySettings settings = MySettings.Load();

            // display banner
            Console.WriteLine("SitemapWiz v 1.0 (c) 2022 by Enrico Savazzi, http://www.savazzi.net.");
            Console.WriteLine("Configuration loaded.");

            // create file list and sitemap
            Console.Write("Generating file list and sitemap ... ");
            Sitemap sitemap = new(settings);
            Console.WriteLine("Done.");

            // display menu
            while (true)
            {
                Console.Write(
                    "\ns - show Sitemap\n" 
                    + "f - show File list\n"
                    + "d - save Default configuration\n"
                    + "r - Reload configuration\n"
                    + "c - show Current configuration\n"
                    + "w - Write sitemap XML file\n"
                    + "q - Quit\n"
                    + "> ");
                ConsoleKeyInfo k = Console.ReadKey();
                switch (k.KeyChar)
                {
                    case 'S':
                        goto case 's';
                    case 's':
                        Console.WriteLine("\n\n-- start Sitemap --\n");
                        Console.Write(sitemap.GetSitemap());
                        Console.WriteLine("\n-- Found "
                            + sitemap.GetFoundURLs()
                            + " URLs and "
                            + sitemap.GetFoundFolders()
                            + " folders, indexed "
                            + sitemap.GetIndexedURLs()
                            + " files in "
                            + sitemap.GetIndexedFolders()
                            + " folders. --");
                        Console.WriteLine("-- end Sitemap --\n\n");
                        break;
                    case 'F':
                        goto case 'f';
                    case 'f':
                        Console.WriteLine("\n\n-- start File List --\n");
                        // sitemap.ProcessBaseFolder(settings.RootFolder, settings.depth, settings.ExcludeFolders.ToArray(), settings.IncludeExtensions.ToArray());
                        Console.Write(sitemap.GetFilelist());
                        Console.WriteLine("\n-- Found " 
                            + sitemap.GetFoundURLs() 
                            + " URLs and " 
                            + sitemap.GetFoundFolders() 
                            + " folders, indexed " 
                            + sitemap.GetIndexedURLs() 
                            + " files in " 
                            + sitemap.GetIndexedFolders() 
                            + " folders. --");
                        Console.WriteLine("-- end File List --\n\n");
                        break;
                    case 'D':
                        goto case 'd';
                    case 'd':
                        MySettings defaultSettings = new MySettings();
                        defaultSettings.Save();
                        Console.WriteLine("\n-- saved default configuration (current configuration is unchanged) --\n\n");
                        break;
                    case 'R':
                        goto case 'r';
                    case 'r':
                        settings = MySettings.Load();
                        Console.WriteLine("\n-- reloaded configuration from configuration file --\n\n");
                        break;
                    case 'C':
                        goto case 'c';
                    case 'c':
                        settings = MySettings.Load();
                        Console.WriteLine("\n-- current configuration start --\n\n");
                        Console.WriteLine(settings.ToString());
                        Console.WriteLine("\n-- current configuration end --\n\n");
                        break;
                    case 'W':
                        goto case 'w';
                    case 'w':
                        Console.Write("\n\n-- writing sitemap ... ");
                        Sitemap.SaveSitemapAsync(settings, sitemap.GetSitemap());
                        Console.WriteLine("Done --\n\n");
                        break;
                    case 'Q':
                        goto case 'q';
                    case 'q':
                        Environment.Exit(1); // exit
                        break;
                    default:
                        Console.WriteLine("\n  Huh???");
                        break;

                }
            }
        }
    }
}

/// <summary>
/// Encapsulates the program settings
/// </summary>
public class MySettings : AppSettings<MySettings>
{
    // settings file name
    public const string DEFAULT_FILENAME = @"SitemapWizSettings.json";
 
    // online site URL
    public string SiteURL = "http://www.savazzi.net/";
    
    // root folder of offline copy of site
    public string RootFolder = @"C:\Users\Enrico\Documents\freehostia\savazzi.net";

    // absolute path and file name of sitemap
    public string SitemapFileName = @"C:\Users\Enrico\Documents\freehostia\savazzi.net\sitemap.xml";

    // depth of inspection into sub-folders
    public int depth = 6;
    
    // include these file extensions in sitemap (files with other extensions are not included)
    public List<string> IncludeExtensions = new() { ".htm", ".html", ".pdf", ".txt", ".rar", ".zip" };
    
    // folders to exclude from sitemap
    public List<string> ExcludeFolders = new() { "images", "img", "scripts", "js", "css", "documents", "reprints" };
    
    // files to exclude from sitemap. Note: case sensitive.
    public List<string> ExcludeFiles = new() { "google6b47bd2ebebf8191.html", "fsg2c635.txt", "robots.txt" };

    public override string ToString()
    {
        string extensions = "";
        foreach (string ext in IncludeExtensions)
        {
            extensions += "        \"" + ext + "\",\n";
        }
        extensions = extensions.Remove(extensions.Length - 2, 2) + "\n";

        string folders = "";
        foreach (string fol in ExcludeFolders)
        {
            folders += "        \"" + fol + "\",\n";
        }
        folders = folders.Remove(folders.Length - 2, 2) + "\n";

        string files = "";
        foreach (string fil in ExcludeFiles)
        {
            files += "        \"" + fil + "\",\n";
        }
        files = files.Remove(files.Length - 2, 2) + "\n";


        string config = "{\n"
            + "    \"DEFAULT_FILENAME\" : \"" + DEFAULT_FILENAME + "\",\n"
            + "    \"SiteURL\" : \"" + SiteURL + "\",\n"
            + "    \"RootFolder\" : \"" + RootFolder + "\",\n"
            + "    \"SitemapFileName\" : \"" + SitemapFileName + "\",\n"
            + "    \"depth\" : \"" + depth.ToString() + "\",\n"
            + "    \"IncludeExtensions\" : [\n" + extensions + "    ]\n"
            + "    \"ExcludeFolders\" : [\n" + folders + "    ]\n"
            + "    \"ExcludeFiles\" : [\n" + files + "    ]\n"
            + "}\n";
        return config;
    }
}

/// <summary>
/// Template of class to encapsulate the program settings
/// </summary>
/// <typeparam name="T">The program settings</typeparam>
public class AppSettings<T> where T : new()
{
    /// <summary>
    /// saves the settings to the specified path\file (in the app folder by default)
    /// </summary>
    /// <param name="fileName">name of settings file</param>
    public void Save(string fileName = MySettings.DEFAULT_FILENAME)
    {
        File.WriteAllText(fileName, new JavaScriptSerializer().Serialize(this));
    }

    /// <summary>
    /// saves the settings to the specified path\file (in the app folder by default)
    /// </summary>
    /// <param name="pSettings">the settings to save</param>
    /// <param name="fileName">name of settings file</param>
    public static void Save(T pSettings, string fileName = MySettings.DEFAULT_FILENAME)
    {
        File.WriteAllText(fileName, new JavaScriptSerializer().Serialize(pSettings));
    }

    /// <summary>
    /// loads the settings from the specified path\file (in the app folder by default)
    /// </summary>
    /// <param name="fileName">name of settings file</param>
    /// <returns>the settings</returns>
    public static T Load(string fileName = MySettings.DEFAULT_FILENAME)
    {
        T t = new T();
        if (File.Exists(fileName))
            t = (new JavaScriptSerializer()).Deserialize<T>(File.ReadAllText(fileName));
        return t;
    }
}

/// <summary>
/// Encapsulates the sitemap
/// </summary>
public class Sitemap
{
    private readonly string urlStart = "  <url>\n", urlEnd = "  </url>\n";
    private readonly string locStart = "    <loc>", locEnd = "</loc>\n";
    private readonly string lastmodStart = "    <lastmod>", lastmodEnd = "</lastmod>\n";
    private readonly string changefreqStart = "    <changefreq>", changefreqEnd = "</changefreq>\n";
    private readonly string priorityStart = "    <priority>", priorityEnd = "</priority>\n";
    private readonly string fileEnd = "</urlset>\n";

    // contains the XML sitemap
    private readonly StringBuilder sitemap = new("", 100000);

    // contains the local file list
    private readonly StringBuilder fileList = new("", 100000);

    // keep count of number of found and indexed files and URLs
    private int foundURLs = 0, indexedURLs = 0, foundFolders = 0, indexedFolders = 0;
    
    // app settings
    private readonly MySettings settings;

    // kep track of whether sitemap is complete
    private bool isComplete = false;


    /// <summary>
    /// constructor
    /// </summary>
    /// <param name="settings">settings used to build sitemap</param>
    public Sitemap(MySettings settings)
    {
        this.settings = settings;
        ProcessBaseFolder(settings.RootFolder, settings.depth, settings.ExcludeFolders.ToArray(), settings.IncludeExtensions.ToArray());
    }

    /// <summary>
    /// Adds one URL to the sitemap
    /// </summary>
    /// <param name="url">The URL to add</param>
    /// <param name="date">The date/time of the file according to the filesystem</param>
    public void AddUrl(string url, DateTime date)
    {
        if (isComplete == false)
        {
            // set the priority to 0.3 by default
            float priority = 0.3f;
            string changefreq = "yearly";

            // 0.5 if file was modified < 1 year ago
            if ((DateTime.Now - date).TotalDays < 365)
            {
                priority = 0.5f;
                changefreq = "monthly";
            }

            // 0.6 if file was modified < 30 days ago
            if ((DateTime.Now - date).TotalDays < 30)
            {
                priority = 0.6f;
                changefreq = "weekly";
            }

            // 0.7 if modified < 7 days ago
            if ((DateTime.Now - date).TotalDays < 7)
            {
                priority = 0.7f;
                changefreq = "daily";
            }

            // 0.9 if modified < 1 day ago
            if ((DateTime.Now - date).TotalHours < 24)
            {
                priority = 0.9f;
                changefreq = "hourly";
            }


            // add the URL to the sitemap
            StringBuilder entry = new(urlStart, 1000);
            entry.Append(locStart + url + locEnd
                + lastmodStart + date.ToString("yyyy-MM-dd") + lastmodEnd
                + priorityStart + priority.ToString() + priorityEnd
                + changefreqStart + changefreq + changefreqEnd
                + urlEnd);
            sitemap.Append(entry);
        }
    }

    public int GetFoundURLs()
    {
        return foundURLs;
    }

    public int GetIndexedURLs()
    {
        return indexedURLs;
    }

    public int GetFoundFolders()
    {
        return foundFolders;
    }

    public int GetIndexedFolders()
    {
        return indexedFolders;
    }

    /// <summary>
    /// Lists the files contained in the specified folder and its subfolders, up to a maximum depth. 
    /// Only files with "included" extensions and folders not "excluded" are listed.
    /// This method calls itself recursively until depth is reached, or no more subfolders remain.
    /// </summary>
    /// <param name="baseFolder">The local folder at the root of the offline copy of the website</param>
    /// <param name="depth">The maximum depth of sub-folders to explore</param>
    /// <param name="excludedFolders">folders to exclude from sitemap</param>
    /// <param name="includedFileExtensions">file extensions to include in sitemap</param>
    public void ProcessBaseFolder(string baseFolder, int depth, string[]? excludedFolders = null, string[]? includedFileExtensions = null)
    {
        // convert arrays to lowercase
        excludedFolders ??= Array.Empty<string>();
        for (int i = 0; i < excludedFolders.Length; i++)
        {
            excludedFolders[i] = excludedFolders[i].ToLower();
        }
        includedFileExtensions ??= Array.Empty<string>();
        for (int i = 0; i < includedFileExtensions.Length; i++)
        {
            includedFileExtensions[i] = includedFileExtensions[i].ToLower();
        }

        // process each filename
        string[] files = Directory.GetFiles(baseFolder);

        // keep totals
        foundURLs += files.Length;
        foreach (string file in files)
        {
            ProcessFile(file);
        }

        foreach (string dir in Directory.GetDirectories(baseFolder))
        {
            foundFolders++;
            if (depth > 0)
            {
                if (Array.IndexOf(excludedFolders, Path.GetFileName(dir)) < 0)
                {
                    indexedFolders++;
                    ProcessBaseFolder(dir, depth - 1, excludedFolders);
                }
            }
        }
    }

    /// <summary>
    /// Processes a file name, adds its URL to the list if it matches all criteria.
    /// </summary>
    /// <param name="file">The file name</param>
    void ProcessFile(string file)
    {
        // get last modified date/time
        DateTime lastModified = System.IO.File.GetLastWriteTime(file);


        // get file extension and convert it to lowercase
        string fileExtension = Path.GetExtension(file).ToLower();

        bool isIncluded = false;

        // validate the file extension
        foreach (string includedExtension in settings.IncludeExtensions)
        {
            if (fileExtension.CompareTo(includedExtension) == 0)
                isIncluded = true;
            if (isIncluded == true)
                break;
        }

        // check if file name is excluded. Note: case sensitive.
        if(isIncluded == true)
        {
            string fileName = Path.GetFileName(file);
            foreach (string excludedFile in settings.ExcludeFiles)
            {
                if (fileName.CompareTo(excludedFile) == 0)
                    isIncluded = false;
                if (isIncluded == false)
                    break;
            }
        }

        if (isIncluded)
        {
            string URL = settings.SiteURL + Path.GetRelativePath(settings.RootFolder, Path.GetFullPath(file));
            URL = URL.Replace('\\', '/');
            URL = URL.Replace(" ", "%20");
            fileList.Append(URL + "\n");

            // add the URL to the sitemap
            AddUrl(URL, lastModified);
            indexedURLs++;
        }

    }

    /// <summary>
    /// returns the sitemap
    /// </summary>
    /// <returns>the sitemap</returns>
    public string GetSitemap()
    {
        if (isComplete == false)
        {
            // add header and root folder to sitemap (we cannot do it earlier because it is only now that we know the URL totals)
            string sDate = DateTime.Now.ToString("yyyy-MM-dd");
            foundURLs++;
            indexedURLs++;

            sitemap.Insert(0, "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n"
                + "<!-- Sitemap for " + settings.SiteURL + " generated at " + sDate + " -->\n"
                + "<!-- " + indexedURLs + " URLs in " + indexedFolders + " folders match the indexing criteria -->\n"
                + "<!-- " + (foundURLs - indexedURLs) + " additional URLs do not match the indexing criteria -->\n"
                + "<!-- " + (foundFolders - indexedFolders) + " folders do not match the indexing criteria and were not crawled -->\n"
                + "<!-- Generated with SitemapWiz v 1.00 (c) by Enrico Savazzi, 2022 -->\n"
                + "<urlset xmlns = \"http://www.sitemaps.org/schemas/sitemap/0.9\" >\n"
                + urlStart
                    + locStart + settings.SiteURL + locEnd
                    + lastmodStart + sDate + lastmodEnd
                    + priorityStart + "1.0" + priorityEnd
                    + changefreqStart + "weekly" + changefreqEnd
                + urlEnd);

            // complete the sitemap
            sitemap.Append(fileEnd);
            isComplete = true;
        }

        return sitemap.ToString();
    }

    public string GetFilelist()
    {
        return fileList.ToString(); 
    }

    public static async void SaveSitemapAsync(MySettings settings, string sitemap)
    {
        await File.WriteAllTextAsync(settings.SitemapFileName, sitemap);
    }
}